@@ -6,7 +6,6 @@ use gimli::{
 use super::{arch::ArchAarch64, unwind_rule::UnwindRuleAarch64, unwindregs::UnwindRegsAarch64};
 
 use crate::unwind_result::UnwindResult;
-use crate::FrameAddress;
 
 use crate::dwarf::{
     eval_cfa_rule, eval_register_rule, ConversionError, DwarfUnwindRegs, DwarfUnwinderError,
@@ -29,7 +28,7 @@ impl DwarfUnwinding for ArchAarch64 {
         unwind_info: &UnwindTableRow<R, S>,
         encoding: Encoding,
         regs: &mut Self::UnwindRegs,
-        address: FrameAddress,
+        is_first_frame: bool,
         read_stack: &mut F,
     ) -> Result<UnwindResult<Self::UnwindRule>, DwarfUnwinderError>
     where
@@ -56,7 +55,7 @@ impl DwarfUnwinding for ArchAarch64 {
         let fp = regs.fp();
         let sp = regs.sp();
 
-        let (fp, lr) = if address.is_return_address() {
+        let (fp, lr) = if !is_first_frame {
             if cfa <= sp {
                 return Err(DwarfUnwinderError::StackPointerMovedBackwards);
             }
@@ -87,9 +86,6 @@ impl DwarfUnwinding for ArchAarch64 {
                 .unwrap_or(fp);
             let lr = eval_register_rule::<R, F, _, S>(lr_rule, cfa, encoding, lr, regs, read_stack)
                 .unwrap_or(lr);
-            if cfa == sp && lr == address.address() {
-                return Err(DwarfUnwinderError::DidNotAdvance);
-            }
             (fp, lr)
         };
 
@@ -125,7 +121,26 @@ fn translate_into_unwind_rule<R: gimli::Reader>(
                 let fp_cfa_offset = register_rule_to_cfa_offset(fp_rule)?;
                 match (lr_cfa_offset, fp_cfa_offset) {
                     (None, Some(_)) => Err(ConversionError::RestoringFpButNotLr),
-                    (None, None) => Ok(UnwindRuleAarch64::OffsetSp { sp_offset_by_16 }),
+                    (None, None) => {
+                        if let RegisterRule::Undefined = lr_rule {
+                            // If the return address is undefined, this could have two reasons:
+                            //  - The column for the return address may have been manually set to "undefined"
+                            //    using DW_CFA_undefined. This usually means that the function never returns
+                            //    and can be treated as the root of the stack.
+                            //  - The column for the return may have been omitted from the DWARF CFI table.
+                            //    Per spec (at least as of DWARF >= 3), this means that it should be treated
+                            //    as undefined. But it seems that compilers often do this when they really mean
+                            //    "same value".
+                            // Gimli follows DWARF 3 and does not differentiate between "omitted" and "undefined".
+                            Ok(
+                                UnwindRuleAarch64::OffsetSpIfFirstFrameOtherwiseStackEndsHere {
+                                    sp_offset_by_16,
+                                },
+                            )
+                        } else {
+                            Ok(UnwindRuleAarch64::OffsetSp { sp_offset_by_16 })
+                        }
+                    }
                     (Some(lr_cfa_offset), None) => {
                         let lr_storage_offset_from_sp_by_8 =
                             i16::try_from((offset + lr_cfa_offset) / 8)
